# Copyright 2021 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import os
import pathlib

from .extensions import load_translate_extensions


def translate(
    *,
    package_name,
    interface_files,
    output_format,
    input_format=None,
    include_paths=None,
    output_path=None,
    translators=None
):
    """
    Translate interface definition files from one format to another.

    To do so, this function leverages translation support as provided
    by third-party package extensions.

    Each path to an interface definition file is a relative path optionally
    prefixed by another path followed by a colon ':', against which the first
    relative path is to be resolved.

    The directory structure that these relative paths exhibit will be
    replicated on output (as opposed to the prefix path, which will be
    ignored).

    If no translators are specified, all available ones will be considered.

    :param package_name: name of the package all interface files belong to
    :param interface_files: list of paths to interface definition files
    :param output_format: format to translate interface definition files to
    :param input_format: optional format to assume for all interface
        definition files, deduced from file extensions if not given
    :param include_paths: optional list of paths to include dependency
        interface definition files from
    :param output_path: optional path to directory to hold translated
        interface definition files, defaults to the current working directory
    :param translators: optional list of translators to use
    :returns: list of paths to translated interface definition files
    """
    extensions = load_translate_extensions(
        specs=translators, strict=bool(translators)
    )
    if not extensions:
        raise RuntimeError('No translate extensions found')

    if not input_format:
        interface_files_per_format = collections.defaultdict(list)
        for interface_file in interface_files:
            input_format = os.path.splitext(interface_file)[-1][1:]
            interface_files_per_format[input_format].append(interface_file)
    else:
        interface_files_per_format = {input_format: interface_files}

    if include_paths is None:
        include_paths = []

    if output_path is None:
        output_path = pathlib.Path.cwd()
    else:
        os.makedirs(output_path, exist_ok=True)

    translated_interface_files = []
    for input_format, interface_files in interface_files_per_format.items():
        extension = next((
            extension for extension in extensions
            if extension.input_format == input_format and
            extension.output_format == output_format
        ), None)

        if not extension:
            raise RuntimeError('\n'.join([
                f"Cannot translate the following files to '{output_format}' format:",
                *[f'- {path}' for path in interface_files],
                'No translator found'
            ]))

        translated_interface_files.extend(extension.translate(
            package_name, interface_files, include_paths, output_path))

    return translated_interface_files



# Copyright 2021 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import re

from rosidl_cli.entry_points import load_entry_points

import yaml


logger = logging.getLogger(__name__)


class Extension:
    """A generic extension point."""

    def __init__(self, name):
        self.__name = name

    @property
    def name(self):
        return self.__name


SPECS_PATTERN = re.compile(r'^(\w+)(?:\[(.+)\])?$')


def parse_extension_specification(spec):
    """
    Parse extension specification.

    :param str spec: specification string in
      'name[key0: value0, ...]' or 'name' format.
      Key-value pairs are parsed as YAML dictionaries.
    :returns: a tuple of specification name and
      keyword arguments, if any, as a dict.
    """
    match = SPECS_PATTERN.match(spec)
    if not match:
        raise ValueError(f'{spec} is not a valid spec')
    name = match.group(1)
    kwargs = match.group(2)
    if kwargs is not None:
        try:
            kwargs = yaml.safe_load('{' + kwargs + '}')
        except Exception as e:
            raise ValueError(
                f'{spec} is not a valid spec'
            ) from e
    else:
        kwargs = {}
    return name, kwargs


def load_extensions(group_name, *, specs=None, strict=False):
    """
    Load extensions for a specific group.

    :param str group_name: the name of the extension group
    :param list specs: an optional collection of extension specs
      (see :py:function:`parse_extension_specification` for spec format)
    :param bool strict: whether to raise or warn on error
    :returns: a list of :py:class:`Extension` instances
    :rtype: list
    """
    extensions = []

    if specs is not None:
        kwargs = dict(map(
            parse_extension_specification, specs))
        specs = list(kwargs.keys())
    else:
        kwargs = {}

    for name, factory in load_entry_points(
        group_name, specs=specs, strict=strict
    ).items():
        try:
            extensions.append(factory(name, **kwargs.get(name, {})))
        except Exception as e:
            msg = f"Failed to instantiate extension '{name}' "
            which = kwargs.get(name, 'default')
            msg += f"with '{which}' arguments: {e}"
            if strict:
                raise RuntimeError(msg)
            logger.warning(msg)
    return extensions



# Copyright 2021 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib

from rosidl_cli.command import Command

from .api import translate


class TranslateCommand(Command):
    """Translate interface definition files from one format to another."""

    name = 'translate'

    def add_arguments(self, parser):
        parser.add_argument(
            '-o', '--output-path', metavar='PATH',
            type=pathlib.Path, default=None,
            help=('Path to directory to hold translated interface '
                  "definition files. Defaults to '.'.")
        )
        parser.add_argument(
            '--use', '--translator', metavar='TRANSLATOR',
            dest='translators', action='append', default=[],
            help=('Translator to be used. If none is specified, '
                  'all available ones will be considered.')
        )
        parser.add_argument(
            '--to', '--output-format', required=True,
            metavar='FORMAT', dest='output_format',
            help='Output format for translate interface definition files.'
        )
        parser.add_argument(
            '--from', '--input-format', default=None,
            metavar='FORMAT', dest='input_format',
            help=('Input format for all source interface definition files. '
                  'If not given, file extensions will be used to deduce '
                  'the format of each interface definition file.')
        )
        parser.add_argument(
            '-I', '--include-path', metavar='PATH', type=pathlib.Path,
            dest='include_paths', action='append', default=[],
            help='Paths to include dependency interface definition files from.'
        )
        parser.add_argument(
            'package_name',
            help='Name of the package all interface files belong to')
        parser.add_argument(
            'interface_files', metavar='interface_file', nargs='+',
            help=('Relative path to an interface definition file. '
                  "If prefixed by another path followed by a colon ':', "
                  'path resolution is performed against such path.')
        )

    def main(self, *, args):
        translate(
            package_name=args.package_name,
            interface_files=args.interface_files,
            output_format=args.output_format,
            input_format=args.input_format,
            include_paths=args.include_paths,
            output_path=args.output_path,
            translators=args.translators
        )



# Copyright 2021 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pathlib

from .extensions import load_type_extensions
from .extensions import load_typesupport_extensions


def generate(
    *,
    package_name,
    interface_files,
    include_paths=None,
    output_path=None,
    types=None,
    typesupports=None
):
    """
    Generate source code from interface definition files.

    To do so, this function leverages type representation and type
    support generation support as provided by third-party package
    extensions.

    Each path to an interface definition file is a relative path optionally
    prefixed by another path followed by a colon ':', against which the first
    relative path is to be resolved.

    The directory structure that these relative paths exhibit will be replicated
    on output (as opposed to the prefix path, which will be ignored).

    If no type representation nor type support is specified, all available ones
    will be generated.

    If more than one type representation or type support is generated, the
    name of each will be appended to the given `output_path` to preclude
    name clashes upon writing source code files.

    :param package_name: name of the package to generate source code for
    :param interface_files: list of paths to interface definition files
    :param include_paths: optional list of paths to include dependency
        interface definition files from
    :param output_path: optional path to directory to hold generated
        source code files, defaults to the current working directory
    :param types: optional list of type representations to generate
    :param typesupports: optional list of type supports to generate
    :returns: list of lists of paths to generated source code files,
        one group per type or type support extension invoked
    """
    extensions = []

    unspecific_generation = not types and not typesupports

    if types or unspecific_generation:
        extensions.extend(load_type_extensions(
            specs=types,
            strict=not unspecific_generation))

    if typesupports or unspecific_generation:
        extensions.extend(load_typesupport_extensions(
            specs=typesupports,
            strict=not unspecific_generation))

    if unspecific_generation and not extensions:
        raise RuntimeError('No type nor typesupport extensions were found')

    if include_paths is None:
        include_paths = []

    if output_path is None:
        output_path = pathlib.Path.cwd()
    else:
        os.makedirs(output_path, exist_ok=True)

    if len(extensions) > 1:
        return [
            extension.generate(
                package_name, interface_files, include_paths,
                output_path=output_path / extension.name)
            for extension in extensions
        ]

    return [extensions[0].generate(
        package_name, interface_files,
        include_paths, output_path
    )]
